# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List
from typing_extensions import TypeAlias

from .category_response import CategoryResponse

__all__ = ["CategoryListResponse"]

CategoryListResponse: TypeAlias = List[CategoryResponse]
